@@ -18,12 +18,17 @@
 #define __iscsi_private_h__
 
 #include <stdint.h>
+#include <time.h>
 
 #if defined(WIN32)
 #include <basetsd.h>
 #define ssize_t SSIZE_T
 #endif
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 #ifndef discard_const
 #define discard_const(ptr) ((void *)((intptr_t)(ptr)))
 #endif
@@ -34,6 +39,7 @@
 #define ISCSI_HEADER_SIZE (ISCSI_RAW_HEADER_SIZE	\
   + (iscsi->header_digest == ISCSI_HEADER_DIGEST_NONE?0:ISCSI_DIGEST_SIZE))
 
+#define SMALL_ALLOC_MAX_FREE (128) /* must be power of 2 */
 
 struct iscsi_in_pdu {
 	struct iscsi_in_pdu *next;
@@ -44,40 +50,43 @@ struct iscsi_in_pdu {
 	long long data_pos;
 	unsigned char *data;
 };
-void iscsi_free_iscsi_in_pdu(struct iscsi_in_pdu *in);
-void iscsi_free_iscsi_inqueue(struct iscsi_in_pdu *inqueue);
-
-enum iscsi_initial_r2t {
-	ISCSI_INITIAL_R2T_NO  = 0,
-	ISCSI_INITIAL_R2T_YES = 1
-};
-
-enum iscsi_immediate_data {
-	ISCSI_IMMEDIATE_DATA_NO  = 0,
-	ISCSI_IMMEDIATE_DATA_YES = 1
-};
+void iscsi_free_iscsi_in_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in);
+void iscsi_free_iscsi_inqueue(struct iscsi_context *iscsi, struct iscsi_in_pdu *inqueue);
 
 struct iscsi_context {
-	const char *initiator_name;
-	const char *target_name;
-	const char *target_address;  /* If a redirect */
-	const char *alias;
+	char initiator_name[MAX_STRING_SIZE+1];
+	char target_name[MAX_STRING_SIZE+1];
+	char target_address[MAX_STRING_SIZE+1];  /* If a redirect */
+	char connected_portal[MAX_STRING_SIZE+1];
+	char portal[MAX_STRING_SIZE+1];
+	char alias[MAX_STRING_SIZE+1];
+	char bind_interfaces[MAX_STRING_SIZE+1];
 
-	const char *user;
-	const char *passwd;
+	char user[MAX_STRING_SIZE+1];
+	char passwd[MAX_STRING_SIZE+1];
+	char chap_c[MAX_STRING_SIZE+1];
+
+	char error_string[MAX_STRING_SIZE+1];
 
 	enum iscsi_session_type session_type;
 	unsigned char isid[6];
 	uint32_t itt;
 	uint32_t cmdsn;
+	uint32_t expcmdsn;
+	uint32_t maxcmdsn;
 	uint32_t statsn;
 	enum iscsi_header_digest want_header_digest;
 	enum iscsi_header_digest header_digest;
 
-	char *error_string;
-
 	int fd;
 	int is_connected;
+	int is_corked;
+
+	int tcp_user_timeout;
+	int tcp_keepcnt;
+	int tcp_keepintvl;
+	int tcp_keepidle;
+	int tcp_syncnt;
 
 	int current_phase;
 	int next_phase;
@@ -88,15 +97,17 @@ struct iscsi_context {
 	int login_attempts;
 	int is_loggedin;
 	int is_reconnecting;
+	int bind_interfaces_cnt;
+	int nops_in_flight;
 
 	int chap_a;
 	int chap_i;
-	char *chap_c;
 
 	iscsi_command_cb socket_status_cb;
 	void *connect_data;
 
 	struct iscsi_pdu *outqueue;
+	struct iscsi_pdu *outqueue_current;
 	struct iscsi_pdu *waitpdu;
 
 	struct iscsi_in_pdu *incoming;
@@ -112,7 +123,24 @@ struct iscsi_context {
 	enum iscsi_immediate_data use_immediate_data;
 
 	int lun;
-	const char *portal;
+	int no_auto_reconnect;
+	int reconnect_deferred;
+	int reconnect_max_retries;
+	int pending_reconnect;
+
+	int log_level;
+	iscsi_log_fn log_fn;
+
+	int mallocs;
+	int reallocs;
+	int frees;
+	int smallocs;
+	void* smalloc_ptrs[SMALL_ALLOC_MAX_FREE];
+	int smalloc_free;
+	size_t smalloc_size;
+
+	time_t last_reconnect;
+	int scsi_timeout;
 };
 
 #define ISCSI_PDU_IMMEDIATE		       0x40
@@ -162,38 +190,63 @@ enum iscsi_opcode {
 	ISCSI_PDU_DATA_IN                        = 0x25,
 	ISCSI_PDU_LOGOUT_RESPONSE                = 0x26,
 	ISCSI_PDU_R2T                            = 0x31,
+	ISCSI_PDU_ASYNC_MSG                      = 0x32,
 	ISCSI_PDU_REJECT                         = 0x3f,
-	ISCSI_PDU_NO_PDU	                 = 0xff
+	ISCSI_PDU_NO_PDU                         = 0xff
+};
+
+struct iscsi_scsi_cbdata {
+	iscsi_command_cb          callback;
+	void                     *private_data;
+	struct scsi_task         *task;
 };
 
 struct iscsi_pdu {
 	struct iscsi_pdu *next;
 
-/* There will not be a response to this pdu, so delete it once it is sent on the wire. Dont put it on the wait-queue */
+/* There will not be a response to this pdu, so delete it once it is sent on the wire. Don't put it on the wait-queue */
 #define ISCSI_PDU_DELETE_WHEN_SENT	0x00000001
-/* Dont call the CANCEL callback when the context is destroyed */
-#define ISCSI_PDU_NO_CALLBACK	 	0x00000002
+/* Don't call the CANCEL callback when the context is destroyed */
+#define ISCSI_PDU_NO_CALLBACK		0x00000002
+/* When reconnecting, just drop all these PDUs. Don't re-queue them.
+ * This includes any DATA-OUT PDU as well as all NOPs.
+ */
+#define ISCSI_PDU_DROP_ON_RECONNECT	0x00000004
+/* stop sending after this PDU has been sent */
+#define ISCSI_PDU_CORK_WHEN_SENT	0x00000008
+/* put this immediate delivery PDU in front of outqueue.
+ * This is currently only used for immediate logout requests
+ * as answer to an async logout event. */
+#define ISCSI_PDU_URGENT_DELIVERY	0x00000010
+
+
 	uint32_t flags;
 
 	uint32_t lun;
 	uint32_t itt;
 	uint32_t cmdsn;
+	uint32_t datasn;
 	enum iscsi_opcode response_opcode;
 
 	iscsi_command_cb callback;
 	void *private_data;
 
-	int written;
-	struct iscsi_data outdata; /* Header and Immediate Data */
+	/* Used to track writing the iscsi header to the socket */
+	struct iscsi_data outdata; /* Header for PDU to send */
+	size_t outdata_written;	   /* How much of the header we have written */
+
+	/* Used to track writing the payload data to the socket */
+	uint32_t payload_offset;   /* Offset of payload data to write */
+	uint32_t payload_len;      /* Amount of payload data to write */
+	uint32_t payload_written;  /* How much of the payload we have written */
+
+
 	struct iscsi_data indata;
 
-	struct iscsi_data nidata; /* Non-Immediate Data */
-
-	struct iscsi_scsi_cbdata *scsi_cbdata;
+	struct iscsi_scsi_cbdata scsi_cbdata;
+	time_t scsi_timeout;
 };
 
-void iscsi_free_scsi_cbdata(struct iscsi_scsi_cbdata *scsi_cbdata);
-
 struct iscsi_pdu *iscsi_allocate_pdu(struct iscsi_context *iscsi,
 				     enum iscsi_opcode opcode,
 				     enum iscsi_opcode response_opcode);
@@ -217,7 +270,8 @@ void iscsi_pdu_set_datasn(struct iscsi_pdu *pdu, uint32_t datasn);
 void iscsi_pdu_set_bufferoffset(struct iscsi_pdu *pdu, uint32_t bufferoffset);
 int iscsi_pdu_add_data(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 		       unsigned char *dptr, int dsize);
-int iscsi_queue_pdu(struct iscsi_context *iscsi, struct iscsi_pdu *pdu);
+EXTERN int iscsi_queue_pdu(struct iscsi_context *iscsi, struct iscsi_pdu *pdu);
+
 int iscsi_add_data(struct iscsi_context *iscsi, struct iscsi_data *data,
 		   unsigned char *dptr, int dsize, int pdualignment);
 
@@ -225,6 +279,7 @@ struct scsi_task;
 void iscsi_pdu_set_cdb(struct iscsi_pdu *pdu, struct scsi_task *task);
 
 int iscsi_get_pdu_data_size(const unsigned char *hdr);
+int iscsi_get_pdu_padding_size(const unsigned char *hdr);
 int iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in);
 
 int iscsi_process_login_reply(struct iscsi_context *iscsi,
@@ -254,18 +309,60 @@ int iscsi_process_reject(struct iscsi_context *iscsi,
 				struct iscsi_in_pdu *in);
 int iscsi_send_target_nop_out(struct iscsi_context *iscsi, uint32_t ttt);
 
+#if defined(WIN32)
 void iscsi_set_error(struct iscsi_context *iscsi, const char *error_string,
 		     ...);
+#else
+void iscsi_set_error(struct iscsi_context *iscsi, const char *error_string,
+		     ...) __attribute__((format(printf, 2, 3)));
+#endif
 
-unsigned char *iscsi_get_user_in_buffer(struct iscsi_context *iscsi, struct iscsi_in_pdu *in, uint32_t pos, ssize_t *count);
-unsigned char *scsi_task_get_data_in_buffer(struct scsi_task *task, uint32_t pos, ssize_t *count);
+struct scsi_iovector *iscsi_get_scsi_task_iovector_in(struct iscsi_context *iscsi, struct iscsi_in_pdu *in);
+struct scsi_iovector *iscsi_get_scsi_task_iovector_out(struct iscsi_context *iscsi, struct iscsi_pdu *pdu);
 
+void* iscsi_malloc(struct iscsi_context *iscsi, size_t size);
+void* iscsi_zmalloc(struct iscsi_context *iscsi, size_t size);
+void* iscsi_realloc(struct iscsi_context *iscsi, void* ptr, size_t size);
+void iscsi_free(struct iscsi_context *iscsi, void* ptr);
+char* iscsi_strdup(struct iscsi_context *iscsi, const char* str);
+void* iscsi_szmalloc(struct iscsi_context *iscsi, size_t size);
+void iscsi_sfree(struct iscsi_context *iscsi, void* ptr);
 
 unsigned long crc32c(char *buf, int len);
 
 struct scsi_task *iscsi_scsi_get_task_from_pdu(struct iscsi_pdu *pdu);
 
-int iscsi_reconnect(struct iscsi_context *iscsi);
+void iscsi_set_noautoreconnect(struct iscsi_context *iscsi, int state);
+
+void iscsi_decrement_iface_rr(void);
+
+#define ISCSI_LOG(iscsi, level, format, ...) \
+	do { \
+		if (level <= iscsi->log_level && iscsi->log_fn) { \
+			iscsi_log_message(iscsi, level, format, ## __VA_ARGS__); \
+		} \
+	} while (0)
+
+void
+iscsi_log_message(struct iscsi_context *iscsi, int level, const char *format, ...);
+
+void
+iscsi_add_to_outqueue(struct iscsi_context *iscsi, struct iscsi_pdu *pdu);
+
+int iscsi_serial32_compare(uint32_t s1, uint32_t s2);
+void iscsi_adjust_statsn(struct iscsi_context *iscsi, struct iscsi_in_pdu *in);
+void iscsi_adjust_maxexpcmdsn(struct iscsi_context *iscsi, struct iscsi_in_pdu *in);
+
+uint32_t iscsi_itt_post_increment(struct iscsi_context *iscsi);
+
+void iscsi_timeout_scan(struct iscsi_context *iscsi);
+
+int
+iscsi_logout_async_internal(struct iscsi_context *iscsi, iscsi_command_cb cb,
+		  void *private_data, uint32_t flags);
+
+#ifdef __cplusplus
+}
+#endif
 
 #endif /* __iscsi_private_h__ */
-
