@@ -18,17 +18,30 @@
 #define __iscsi_h__
 
 #include <stdint.h>
+#include <sys/types.h>
 
 #if defined(WIN32)
 #define EXTERN __declspec( dllexport )
-EXTERN int poll(struct pollfd *fds, int nfsd, int timeout);
 #else
 #define EXTERN
 #endif
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 struct iscsi_context;
 struct sockaddr;
 
+/* API VERSION */
+#define LIBISCSI_API_VERSION (20140713)
+
+/* FEATURES */
+#define LIBISCSI_FEATURE_IOVECTOR (1)
+#define LIBISCSI_FEATURE_NOP_COUNTER (1)
+
+#define MAX_STRING_SIZE (255)
+
 /*
  * Syntax for normal and portal/discovery URLs.
  */
@@ -61,26 +74,64 @@ EXTERN int iscsi_service(struct iscsi_context *iscsi, int revents);
 EXTERN int iscsi_queue_length(struct iscsi_context *iscsi);
 
 /*
- * To set tcp keepalive for the session
- * ONLY available on platforms with tcp keepalive socket options
+ * Set the timeout in seconds after which a synchronous SCSI command
+ * will timeout.
+ * Default is 0 == no timeout.
  */
-int iscsi_set_tcp_keepalive(struct iscsi_context *iscsi, int idle, int count, int interval);
-
+EXTERN int iscsi_set_timeout(struct iscsi_context *iscsi, int timeout);
 
+/*
+ * To set tcp keepalive for the session.
+ * Only options supported by given platform (if any) are set.
+ */
+EXTERN int iscsi_set_tcp_keepalive(struct iscsi_context *iscsi, int idle, int count, int interval);
 
 struct iscsi_url {
-       const char *portal;
-       const char *target;
-       const char *user;
-       const char *passwd;
+       char portal[MAX_STRING_SIZE + 1];
+       char target[MAX_STRING_SIZE + 1];
+       char user[MAX_STRING_SIZE + 1];
+       char passwd[MAX_STRING_SIZE + 1];
        int lun;
+       struct iscsi_context *iscsi;
 };
 
+/*
+ * This function is used to set the desired mode for immediate data.
+ * This can be set on a context before it has been logged in to the target
+ * and controls how the initiator will try to negotiate the immediate data.
+ *
+ * Default is for libiscsi to try to negotiate ISCSI_IMMEDIATE_DATA_YES
+ */
+enum iscsi_immediate_data {
+	ISCSI_IMMEDIATE_DATA_NO  = 0,
+	ISCSI_IMMEDIATE_DATA_YES = 1
+};
+int iscsi_set_immediate_data(struct iscsi_context *iscsi, enum iscsi_immediate_data immediate_data);
+
+/*
+ * This function is used to set the desired mode for initial_r2t
+ * This can be set on a context before it has been logged in to the target
+ * and controls how the initiator will try to negotiate the initial r2t.
+ *
+ * Default is for libiscsi to try to negotiate ISCSI_INITIAL_R2T_NO
+ */
+enum iscsi_initial_r2t {
+	ISCSI_INITIAL_R2T_NO  = 0,
+	ISCSI_INITIAL_R2T_YES = 1
+};
+int
+iscsi_set_initial_r2t(struct iscsi_context *iscsi, enum iscsi_initial_r2t initial_r2t);
+
+
 /*
  * This function is used to parse an iSCSI URL into a iscsi_url structure.
  * iSCSI URL format :
  * iscsi://[<username>[%<password>]@]<host>[:<port>]/<target-iqn>/<lun>
  *
+ * Target names are url encoded with '%' as a special character.
+ * Example:
+ * "iqn.ronnie.test%3A1234" will be translated to "iqn.ronnie.test:1234"
+ *
  * Function will return a pointer to an iscsi url structure if successful,
  * or it will return NULL and set iscsi_get_error() accrodingly if there was a problem
  * with the URL.
@@ -227,9 +278,16 @@ EXTERN int iscsi_is_logged_in(struct iscsi_context *iscsi);
 enum scsi_status {
 	SCSI_STATUS_GOOD                 = 0,
 	SCSI_STATUS_CHECK_CONDITION      = 2,
+	SCSI_STATUS_CONDITION_MET        = 4,
+	SCSI_STATUS_BUSY                 = 8,
 	SCSI_STATUS_RESERVATION_CONFLICT = 0x18,
+	SCSI_STATUS_TASK_SET_FULL        = 0x28,
+	SCSI_STATUS_ACA_ACTIVE           = 0x30,
+	SCSI_STATUS_TASK_ABORTED         = 0x40,
+	SCSI_STATUS_REDIRECT             = 0x101,
 	SCSI_STATUS_CANCELLED            = 0x0f000000,
-	SCSI_STATUS_ERROR                = 0x0f000001
+	SCSI_STATUS_ERROR                = 0x0f000001,
+	SCSI_STATUS_TIMEOUT              = 0x0f000002
 };
 
 
@@ -329,6 +387,15 @@ EXTERN int iscsi_full_connect_sync(struct iscsi_context *iscsi, const char *port
  */
 EXTERN int iscsi_disconnect(struct iscsi_context *iscsi);
 
+/*
+ * Disconnect a connection to a target and try to reconnect.
+ *
+ * Returns:
+ *  0 reconnect was successful
+ * <0 error
+ */
+EXTERN int iscsi_reconnect(struct iscsi_context *iscsi);
+
 /*
  * Asynchronous call to perform an ISCSI login.
  *
@@ -407,10 +474,15 @@ EXTERN int iscsi_logout_sync(struct iscsi_context *iscsi);
 EXTERN int iscsi_discovery_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 			  void *private_data);
 
+struct iscsi_target_portal {
+       struct iscsi_target_portal *next;
+       const char *portal;
+};
+
 struct iscsi_discovery_address {
        struct iscsi_discovery_address *next;
        const char *target_name;
-       const char *target_address;
+       struct iscsi_target_portal *portals;
 };
 
 /*
@@ -428,10 +500,17 @@ struct iscsi_discovery_address {
  *                            structure containing the data returned from
  *                            the server.
  *    ISCSI_STATUS_CANCELLED: Discovery was aborted. Command_data is NULL.
+ * 
+ * The callback may be NULL if you only want to let libiscsi count the in-flight
+ * NOPs.
  */
 EXTERN int iscsi_nop_out_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 			unsigned char *data, int len, void *private_data);
 
+
+/* read out the number of consecutive nop outs that did not receive an answer */
+EXTERN int iscsi_get_nops_in_flight(struct iscsi_context *iscsi);
+
 struct scsi_task;
 
 enum iscsi_task_mgmt_funcs {
@@ -466,11 +545,11 @@ enum iscsi_task_mgmt_funcs {
  * abort_task will also cancel the scsi task. The callback for the scsi task will be invoked with
  *            SCSI_STATUS_CANCELLED
  * abort_task_set, lun_reset, target_warn_reset, target_cold_reset will cancel all tasks. The callback for
- *            all tasks will be invoked with SCSI_STATUS_CANCELLED 
+ *            all tasks will be invoked with SCSI_STATUS_CANCELLED
  */
 EXTERN int
 iscsi_task_mgmt_async(struct iscsi_context *iscsi,
-		      int lun, enum iscsi_task_mgmt_funcs function, 
+		      int lun, enum iscsi_task_mgmt_funcs function,
 		      uint32_t ritt, uint32_t rcmdscn,
 		      iscsi_command_cb cb, void *private_data);
 
@@ -495,6 +574,33 @@ iscsi_task_mgmt_target_cold_reset_async(struct iscsi_context *iscsi,
 
 
 
+/*
+ * Synchronous calls for task management
+ *
+ * Returns:
+ *  0 success.
+ * <0 error.
+ */
+EXTERN int
+iscsi_task_mgmt_sync(struct iscsi_context *iscsi,
+		     int lun, enum iscsi_task_mgmt_funcs function,
+		     uint32_t ritt, uint32_t rcmdscn);
+
+EXTERN int
+iscsi_task_mgmt_abort_task_sync(struct iscsi_context *iscsi, struct scsi_task *task);
+
+EXTERN int
+iscsi_task_mgmt_abort_task_set_sync(struct iscsi_context *iscsi, uint32_t lun);
+
+EXTERN int
+iscsi_task_mgmt_lun_reset_sync(struct iscsi_context *iscsi, uint32_t lun);
+
+EXTERN int
+iscsi_task_mgmt_target_warm_reset_sync(struct iscsi_context *iscsi);
+
+EXTERN int
+iscsi_task_mgmt_target_cold_reset_sync(struct iscsi_context *iscsi);
+
 
 
 
@@ -518,7 +624,7 @@ iscsi_task_mgmt_target_cold_reset_async(struct iscsi_context *iscsi,
  */
 
 struct iscsi_data {
-       int size;
+       size_t size;
        unsigned char *data;
 };
 
@@ -541,9 +647,11 @@ EXTERN int
 iscsi_set_isid_reserved(struct iscsi_context *iscsi);
 
 
+struct scsi_mode_page;
 
 
-/* 
+
+/*
  * The scsi commands use/return a scsi_task structure when invoked
  * and also through the callback.
  *
@@ -580,14 +688,41 @@ iscsi_readcapacity10_task(struct iscsi_context *iscsi, int lun, int lba,
 			       void *private_data);
 EXTERN struct scsi_task *
 iscsi_readcapacity16_task(struct iscsi_context *iscsi, int lun,
-			       iscsi_command_cb cb,
-			       void *private_data);
+			  iscsi_command_cb cb,
+			  void *private_data);
+EXTERN struct scsi_task *
+iscsi_sanitize_task(struct iscsi_context *iscsi, int lun,
+		    int immed, int ause, int sa, int param_len,
+		    struct iscsi_data *data,
+		    iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_sanitize_block_erase_task(struct iscsi_context *iscsi, int lun,
+			       int immed, int ause,
+			       iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_sanitize_crypto_erase_task(struct iscsi_context *iscsi, int lun,
+				 int immed, int ause,
+				 iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_sanitize_exit_failure_mode_task(struct iscsi_context *iscsi, int lun,
+				      int immed, int ause,
+				      iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_get_lba_status_task(struct iscsi_context *iscsi, int lun,
+			  uint64_t starting_lba, uint32_t alloc_len,
+			  iscsi_command_cb cb,
+			  void *private_data);
 EXTERN struct scsi_task *
 iscsi_synchronizecache10_task(struct iscsi_context *iscsi, int lun,
 				   int lba, int num_blocks, int syncnv,
 				   int immed, iscsi_command_cb cb,
 				   void *private_data);
 EXTERN struct scsi_task *
+iscsi_synchronizecache16_task(struct iscsi_context *iscsi, int lun,
+				   uint64_t lba, uint32_t num_blocks, int syncnv,
+				   int immed, iscsi_command_cb cb,
+				   void *private_data);
+EXTERN struct scsi_task *
 iscsi_prefetch10_task(struct iscsi_context *iscsi, int lun,
 		      uint32_t lba, int num_blocks,
 		      int immed, int group,
@@ -604,69 +739,156 @@ iscsi_read6_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 
 EXTERN struct scsi_task *
 iscsi_read10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
-		       uint32_t datalen, int blocksize, iscsi_command_cb cb,
-		       void *private_data);
+		  uint32_t datalen, int blocksize,
+		  int rdprotect, int dpo, int fua, int fua_nv, int group_number,
+		  iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
-iscsi_write10_task(struct iscsi_context *iscsi, int lun,
-			unsigned char *data, uint32_t datalen, uint32_t lba, int fua,
-			int fua_nv, int blocksize, iscsi_command_cb cb,
-			void *private_data);
+iscsi_write10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_writeverify10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int bytchk, int group_number,
+		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
 iscsi_read12_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   uint32_t datalen, int blocksize,
 		   int rdprotect, int dpo, int fua, int fua_nv, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
-iscsi_write12_task(struct iscsi_context *iscsi, int lun, uint32_t lba, 
+iscsi_write12_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_writeverify12_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int bytchk, int group_number,
+		   iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
 iscsi_read16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   uint32_t datalen, int blocksize,
 		   int rdprotect, int dpo, int fua, int fua_nv, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
-iscsi_write16_task(struct iscsi_context *iscsi, int lun, uint64_t lba, 
+iscsi_write16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_orwrite_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_startstopunit_task(struct iscsi_context *iscsi, int lun,
+			 int immed, int pcm, int pc,
+			 int no_flush, int loej, int start,
+			 iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_preventallow_task(struct iscsi_context *iscsi, int lun,
+			int prevent,
+			iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_compareandwrite_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_writeverify16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int bytchk, int group_number,
+		   iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
 iscsi_verify10_task(struct iscsi_context *iscsi, int lun,
 		    unsigned char *data, uint32_t datalen, uint32_t lba,
-		    int vprotect, int dpo, int bytchk, 
+		    int vprotect, int dpo, int bytchk,
 		    int blocksize, iscsi_command_cb cb,
 		    void *private_data);
 EXTERN struct scsi_task *
-iscsi_writesame10_task(struct iscsi_context *iscsi, int lun,
+iscsi_verify12_task(struct iscsi_context *iscsi, int lun,
+		    unsigned char *data, uint32_t datalen, uint32_t lba,
+		    int vprotect, int dpo, int bytchk,
+		    int blocksize, iscsi_command_cb cb,
+		    void *private_data);
+EXTERN struct scsi_task *
+iscsi_verify16_task(struct iscsi_context *iscsi, int lun,
+		    unsigned char *data, uint32_t datalen, uint64_t lba,
+		    int vprotect, int dpo, int bytchk,
+		    int blocksize, iscsi_command_cb cb,
+		    void *private_data);
+EXTERN struct scsi_task *
+iscsi_writesame10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		       unsigned char *data, uint32_t datalen,
-		       uint32_t lba, uint16_t num_blocks,
-		       int anchor, int unmap, int pbdata, int lbdata,
-		       int wrprotect, int group,
+		       uint16_t num_blocks,
+		       int anchor, int unmap, int wrprotect, int group,
 		       iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
-iscsi_writesame16_task(struct iscsi_context *iscsi, int lun,
+iscsi_writesame16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		       unsigned char *data, uint32_t datalen,
-		       uint64_t lba, uint32_t num_blocks,
-		       int anchor, int unmap, int pbdata, int lbdata,
-		       int wrprotect, int group,
+		       uint32_t num_blocks,
+		       int anchor, int unmap, int wrprotect, int group,
 		       iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_modeselect6_task(struct iscsi_context *iscsi, int lun,
+		       int pf, int sp, struct scsi_mode_page *mp,
+		       iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_modeselect10_task(struct iscsi_context *iscsi, int lun,
+			int pf, int sp, struct scsi_mode_page *mp,
+			iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
 iscsi_modesense6_task(struct iscsi_context *iscsi, int lun, int dbd,
 			   int pc, int page_code, int sub_page_code,
 			   unsigned char alloc_len, iscsi_command_cb cb,
 			   void *private_data);
+EXTERN struct scsi_task *
+iscsi_modesense10_task(struct iscsi_context *iscsi, int lun, int llbaa, int dbd,
+			   int pc, int page_code, int sub_page_code,
+			   unsigned char alloc_len, iscsi_command_cb cb,
+			   void *private_data);
 
 struct unmap_list {
        uint64_t	  lba;
        uint32_t	  num;
 };
 
+EXTERN struct scsi_task *
+iscsi_persistent_reserve_in_task(struct iscsi_context *iscsi, int lun,
+				 int sa, uint16_t xferlen,
+				 iscsi_command_cb cb, void *private_data);
+
+EXTERN struct scsi_task *
+iscsi_persistent_reserve_out_task(struct iscsi_context *iscsi, int lun,
+				  int sa, int scope, int type, void *params,
+				  iscsi_command_cb cb, void *private_data);
+
 EXTERN struct scsi_task *
 iscsi_unmap_task(struct iscsi_context *iscsi, int lun, int anchor, int group,
 		 struct unmap_list *list, int list_len,
 		 iscsi_command_cb cb, void *private_data);
 
+EXTERN struct scsi_task *
+iscsi_readtoc_task(struct iscsi_context *iscsi, int lun, int msf, int format,
+		   int track_session, int maxsize,
+		   iscsi_command_cb cb, void *private_data);
+
+EXTERN struct scsi_task *
+iscsi_reserve6_task(struct iscsi_context *iscsi, int lun,
+		    iscsi_command_cb cb, void *private_data);
+
+EXTERN struct scsi_task *
+iscsi_release6_task(struct iscsi_context *iscsi, int lun,
+		    iscsi_command_cb cb, void *private_data);
+
+EXTERN struct scsi_task *
+iscsi_report_supported_opcodes_task(struct iscsi_context *iscsi, int lun,
+				    int rctd, int options,
+				    int opcode, int sa,
+				    uint32_t alloc_len,
+				    iscsi_command_cb cb, void *private_data);
 
 /*
  * Sync commands for SCSI
@@ -675,6 +897,24 @@ EXTERN struct scsi_task *
 iscsi_scsi_command_sync(struct iscsi_context *iscsi, int lun,
 			struct scsi_task *task, struct iscsi_data *data);
 
+EXTERN struct scsi_task *
+iscsi_modeselect6_sync(struct iscsi_context *iscsi, int lun,
+		       int pf, int sp, struct scsi_mode_page *mp);
+
+EXTERN struct scsi_task *
+iscsi_modeselect10_sync(struct iscsi_context *iscsi, int lun,
+			int pf, int sp, struct scsi_mode_page *mp);
+
+EXTERN struct scsi_task *
+iscsi_modesense6_sync(struct iscsi_context *iscsi, int lun, int dbd,
+		      int pc, int page_code, int sub_page_code,
+		      unsigned char alloc_len);
+
+EXTERN struct scsi_task *
+iscsi_modesense10_sync(struct iscsi_context *iscsi, int lun, int llbaa, int dbd,
+		      int pc, int page_code, int sub_page_code,
+		      unsigned char alloc_len);
+
 EXTERN struct scsi_task *
 iscsi_reportluns_sync(struct iscsi_context *iscsi, int report_type,
 		      int alloc_len);
@@ -692,12 +932,18 @@ iscsi_read6_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 
 EXTERN struct scsi_task *
 iscsi_read10_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
-		  uint32_t datalen, int blocksize);
+		  uint32_t datalen, int blocksize,
+		  int rdprotect, int dpo, int fua, int fua_nv, int group_number);
 
 EXTERN struct scsi_task *
-iscsi_write10_sync(struct iscsi_context *iscsi, int lun,
-			unsigned char *data, uint32_t datalen, uint32_t lba, int fua,
-			int fua_nv, int blocksize);
+iscsi_write10_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
+
+EXTERN struct scsi_task *
+iscsi_writeverify10_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int bytchk, int group_number);
 
 EXTERN struct scsi_task *
 iscsi_read12_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
@@ -709,6 +955,11 @@ iscsi_write12_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_writeverify12_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int bytchk, int group_number);
+
 EXTERN struct scsi_task *
 iscsi_read16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		  uint32_t datalen, int blocksize,
@@ -719,6 +970,30 @@ iscsi_write16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_orwrite_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
+
+EXTERN struct scsi_task *
+iscsi_startstopunit_sync(struct iscsi_context *iscsi, int lun,
+			 int immed, int pcm, int pc,
+			 int no_flush, int loej, int start);
+
+EXTERN struct scsi_task *
+iscsi_preventallow_sync(struct iscsi_context *iscsi, int lun,
+			int prevent);
+
+EXTERN struct scsi_task *
+iscsi_compareandwrite_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
+
+EXTERN struct scsi_task *
+iscsi_writeverify16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int bytchk, int group_number);
+
 EXTERN struct scsi_task *
 iscsi_readcapacity10_sync(struct iscsi_context *iscsi, int lun, int lba,
 			  int pmi);
@@ -726,10 +1001,30 @@ iscsi_readcapacity10_sync(struct iscsi_context *iscsi, int lun, int lba,
 EXTERN struct scsi_task *
 iscsi_readcapacity16_sync(struct iscsi_context *iscsi, int lun);
 
+EXTERN struct scsi_task *
+iscsi_get_lba_status_sync(struct iscsi_context *iscsi, int lun, uint64_t starting_lba, uint32_t alloc_len);
+
+EXTERN struct scsi_task *
+iscsi_sanitize_sync(struct iscsi_context *iscsi, int lun,
+		    int immed, int ause, int sa, int param_len,
+		    struct iscsi_data *data);
+EXTERN struct scsi_task *
+iscsi_sanitize_block_erase_sync(struct iscsi_context *iscsi, int lun,
+		    int immed, int ause);
+EXTERN struct scsi_task *
+iscsi_sanitize_crypto_erase_sync(struct iscsi_context *iscsi, int lun,
+		    int immed, int ause);
+EXTERN struct scsi_task *
+iscsi_sanitize_exit_failure_mode_sync(struct iscsi_context *iscsi, int lun,
+		    int immed, int ause);
 EXTERN struct scsi_task *
 iscsi_synchronizecache10_sync(struct iscsi_context *iscsi, int lun, int lba,
 			      int num_blocks, int syncnv, int immed);
 
+EXTERN struct scsi_task *
+iscsi_synchronizecache16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			      uint32_t num_blocks, int syncnv, int immed);
+
 EXTERN struct scsi_task *
 iscsi_prefetch10_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		      int num_blocks, int immed, int group);
@@ -741,49 +1036,96 @@ iscsi_prefetch16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
 EXTERN struct scsi_task *
 iscsi_verify10_sync(struct iscsi_context *iscsi, int lun,
 		    unsigned char *data, uint32_t datalen, uint32_t lba,
-		    int vprotect, int dpo, int bytchk, 
+		    int vprotect, int dpo, int bytchk,
 		    int blocksize);
 
 EXTERN struct scsi_task *
-iscsi_writesame10_sync(struct iscsi_context *iscsi, int lun,
-		       unsigned char *data, uint32_t datalen,
-		       uint32_t lba, uint16_t num_blocks,
-		       int anchor, int unmap, int pbdata, int lbdata,
-		       int wrprotect, int group);
+iscsi_verify12_sync(struct iscsi_context *iscsi, int lun,
+		    unsigned char *data, uint32_t datalen, uint32_t lba,
+		    int vprotect, int dpo, int bytchk,
+		    int blocksize);
 
 EXTERN struct scsi_task *
-iscsi_writesame16_sync(struct iscsi_context *iscsi, int lun,
+iscsi_verify16_sync(struct iscsi_context *iscsi, int lun,
+		    unsigned char *data, uint32_t datalen, uint64_t lba,
+		    int vprotect, int dpo, int bytchk,
+		    int blocksize);
+
+EXTERN struct scsi_task *
+iscsi_writesame10_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		       unsigned char *data, uint32_t datalen,
-		       uint64_t lba, uint32_t num_blocks,
-		       int anchor, int unmap, int pbdata, int lbdata,
-		       int wrprotect, int group);
+		       uint16_t num_blocks,
+		       int anchor, int unmap, int wrprotect, int group);
+
+EXTERN struct scsi_task *
+iscsi_writesame16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		       unsigned char *data, uint32_t datalen,
+		       uint32_t num_blocks,
+		       int anchor, int unmap, int wrprotect, int group);
+
+EXTERN struct scsi_task *
+iscsi_persistent_reserve_in_sync(struct iscsi_context *iscsi, int lun,
+				 int sa, uint16_t xferlen);
+
+EXTERN struct scsi_task *
+iscsi_persistent_reserve_out_sync(struct iscsi_context *iscsi, int lun,
+				  int sa, int scope, int type, void *params);
 
 EXTERN struct scsi_task *
 iscsi_unmap_sync(struct iscsi_context *iscsi, int lun, int anchor, int group,
 		 struct unmap_list *list, int list_len);
 
+EXTERN struct scsi_task *
+iscsi_readtoc_sync(struct iscsi_context *iscsi, int lun, int msf,
+		   int format, int track_session, int maxsize);
+
+EXTERN struct scsi_task *
+iscsi_reserve6_sync(struct iscsi_context *iscsi, int lun);
+
+EXTERN struct scsi_task *
+iscsi_release6_sync(struct iscsi_context *iscsi, int lun);
+
+EXTERN struct scsi_task *
+iscsi_report_supported_opcodes_sync(struct iscsi_context *iscsi, int lun,
+				    int rctd, int options,
+				    int opcode, int sa,
+				    uint32_t alloc_len);
+
 /*
- * This function is used when  the application wants to specify its own buffers to read the data
- * from the DATA-IN PDUs into.
- * The main use is for SCSI read operations to have them write directly into the application buffers to
- * avoid the two copies that would occur otherwise.
- * First copy from the individual DATA-IN blobs to linearize the buffer and the second in the callback
- * to copy the data from the linearized buffer into the application buffer.
+ * These functions are used when the application wants to specify its own buffers to read the data
+ * from the DATA-IN PDUs into, or write the data to DATA-OUT PDUs from.
+ * The main use is for SCSI READ10/12/16 WRITE10/12/16 operations to have them read/write directly from
+ * the applications buffer, avoiding coying the data.
  *
  * This also supports reading into a vector of buffers by calling this function multiple times.
  * The individual buffers will be filled in the same order as they were created.
  *
- * Example:
+ * Example READ10:
  *     task = iscsi_read10_task(    ( 2 512byte blocks into two buffers)
  *     scsi_task_add_data_in_buffer(task, first_buffer, 512
  *     scsi_task_add_data_in_buffer(task, second_buffer, 512
  *
  *
- * If you use this function you can not use task->datain in the callback.
+ * If you use this function you can not use task->datain in the READ callback.
  * task->datain.size will be 0 and
  * task->datain.data will be NULL
+ *
+ * Example WRITE10: (write 2 blocks)
+ *     static struct scsi_iovec iov[2];
+ *
+ *     task = iscsi_write10_task(iscsi, lun, 0, NULL, 512, 512, 0, 0, 0, 0, 0, callback, private_data);
+ *     iov[0].iov_base = first_buffer;
+ *     iov[0].iov_len  = 512;
+ *     iov[1].iov_base = second_buffer;
+ *     iov[1].iov_len  = 512;
+ *     scsi_task_set_iov_out(task, &iov[0], 2);
  */
 EXTERN int scsi_task_add_data_in_buffer(struct scsi_task *task, int len, unsigned char *buf);
+EXTERN int scsi_task_add_data_out_buffer(struct scsi_task *task, int len, unsigned char *buf);
+
+struct scsi_iovec;
+EXTERN void scsi_task_set_iov_out(struct scsi_task *task, struct scsi_iovec *iov, int niov);
+EXTERN void scsi_task_set_iov_in(struct scsi_task *task, struct scsi_iovec *iov, int niov);
 
 /*
  * This function is used when you want to cancel a scsi task.
@@ -793,8 +1135,8 @@ EXTERN int scsi_task_add_data_in_buffer(struct scsi_task *task, int len, unsigne
  * To cancel the task also a the target you need to call the task management functions.
  */
 EXTERN int
-iscsi_scsi_task_cancel(struct iscsi_context *iscsi,
-		  struct scsi_task *task);
+iscsi_scsi_cancel_task(struct iscsi_context *iscsi,
+		       struct scsi_task *task);
 
 /*
  * This function is used when you want to cancel all scsi tasks.
@@ -806,4 +1148,82 @@ iscsi_scsi_task_cancel(struct iscsi_context *iscsi,
 EXTERN void
 iscsi_scsi_cancel_all_tasks(struct iscsi_context *iscsi);
 
+/*
+ * This function is to set the debugging level where level is
+ *
+ * 0  = disabled (default)
+ * 1  = errors only
+ * 2  = connection related info
+ * 3  = user set variables
+ * 4  = function calls
+ * 5  = ...
+ * 10 = everything
+ */
+EXTERN void
+iscsi_set_log_level(struct iscsi_context *iscsi, int level);
+
+typedef void (*iscsi_log_fn)(int level, const char *mesage);
+
+/* Set the logging function to use */
+EXTERN void iscsi_set_log_fn(struct iscsi_context *iscsi, iscsi_log_fn fn);
+
+/* predefined log function that just writes to stderr */
+EXTERN void iscsi_log_to_stderr(int level, const char *message);
+
+/*
+ * This function is to set the TCP_USER_TIMEOUT option. It has to be called after iscsi
+ * context creation. The value given in ms is then applied each time a new socket is created.
+ */
+EXTERN void
+iscsi_set_tcp_user_timeout(struct iscsi_context *iscsi, int timeout_ms);
+
+/*
+ * This function is to set the TCP_KEEPIDLE option. It has to be called after iscsi
+ * context creation.
+ */
+EXTERN void
+iscsi_set_tcp_keepidle(struct iscsi_context *iscsi, int value);
+
+/*
+ * This function is to set the TCP_KEEPCNT option. It has to be called after iscsi
+ * context creation.
+ */
+EXTERN void
+iscsi_set_tcp_keepcnt(struct iscsi_context *iscsi, int value);
+
+/*
+ * This function is to set the TCP_KEEPINTVL option. It has to be called after iscsi
+ * context creation.
+ */
+EXTERN void
+iscsi_set_tcp_keepintvl(struct iscsi_context *iscsi, int value);
+
+/*
+ * This function is to set the TCP_SYNCNT option. It has to be called after iscsi
+ * context creation.
+ */
+EXTERN void
+iscsi_set_tcp_syncnt(struct iscsi_context *iscsi, int value);
+
+/*
+ * This function is to set the interface that outbound connections for this socket are bound to.
+ * You max specify more than one interface here separated by comma.
+ */
+EXTERN void
+iscsi_set_bind_interfaces(struct iscsi_context *iscsi, char * interfaces);
+
+/* This function is to set if we should retry a failed reconnect
+   
+   count is defined as follows:
+    -1 -> retry forever (default)
+    0  -> never retry
+    n  -> retry n times
+*/
+EXTERN void
+iscsi_set_reconnect_max_retries(struct iscsi_context *iscsi, int count);
+
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* __iscsi_h__ */
