@@ -19,23 +19,39 @@
 #define _GNU_SOURCE
 #endif
 
-#if defined(WIN32)
-#include <winsock2.h>
-#else
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
+#ifdef HAVE_SYS_TYPES_H
+#include <sys/types.h>
+#endif
+
+#ifdef HAVE_ARPA_INET_H
 #include <arpa/inet.h>
 #endif
 
+#if defined(WIN32)
+#include <winsock2.h>
+#include "win32/win32_compat.h"
+#endif
+
 #include <stdio.h>
+#include <assert.h>
 #include <stdlib.h>
 #include <string.h>
 #include "iscsi.h"
 #include "iscsi-private.h"
+#include "scsi-lowlevel.h"
 #include "md5.h"
+#ifdef HAVE_LIBGCRYPT
+#include <gcrypt.h>
+#endif
 
 static int
 iscsi_login_add_initiatorname(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send InitiatorName during opneg or the first leg of secneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
@@ -43,30 +59,22 @@ iscsi_login_add_initiatorname(struct iscsi_context *iscsi, struct iscsi_pdu *pdu
 		return 0;
 	}
 
-	str = malloc(1024);
-#if defined(WIN32)
-	if (_snprintf_s(str, 1024, 1024, "InitiatorName=%s", iscsi->initiator_name) == -1) {
-#else
-	if (snprintf(str, 1024, "InitiatorName=%s", iscsi->initiator_name) == -1) {
-#endif
+	if (snprintf(str, MAX_STRING_SIZE, "InitiatorName=%s", iscsi->initiator_name) == -1) {
 		iscsi_set_error(iscsi, "Out-of-memory: aprintf failed.");
-		free(str);
 		return -1;
 	}
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
-		free(str);
 		return -1;
 	}
-	free(str);
 	return 0;
 }
 
 static int
 iscsi_login_add_alias(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send InitiatorAlias during opneg or the first leg of secneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
@@ -74,31 +82,23 @@ iscsi_login_add_alias(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 		return 0;
 	}
 
-	str = malloc(1024);
-#if defined(WIN32)
-	if (_snprintf_s(str, 1024, 1024, "InitiatorAlias=%s", iscsi->alias) == -1) {
-#else
-	if (snprintf(str, 1024, "InitiatorAlias=%s", iscsi->alias) == -1) {
-#endif
+	if (snprintf(str, MAX_STRING_SIZE, "InitiatorAlias=%s", iscsi->alias) == -1) {
 		iscsi_set_error(iscsi, "Out-of-memory: aprintf failed.");
-		free(str);
 		return -1;
 	}
 
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
-		free(str);
 		return -1;
 	}
-	free(str);
 	return 0;
 }
 
 static int
 iscsi_login_add_targetname(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send TargetName during opneg or the first leg of secneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
@@ -106,37 +106,29 @@ iscsi_login_add_targetname(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 		return 0;
 	}
 
-	if (iscsi->target_name == NULL) {
+	if (!iscsi->target_name[0]) {
 		iscsi_set_error(iscsi, "Trying normal connect but "
 				"target name not set.");
 		return -1;
 	}
 
-	str = malloc(1024);
-#if defined(WIN32)
-	if (_snprintf_s(str, 1024, 1024, "TargetName=%s", iscsi->target_name) == -1) {
-#else
-	if (snprintf(str, 1024, "TargetName=%s", iscsi->target_name) == -1) {
-#endif
+	if (snprintf(str, MAX_STRING_SIZE, "TargetName=%s", iscsi->target_name) == -1) {
 		iscsi_set_error(iscsi, "Out-of-memory: aprintf failed.");
-		free(str);
 		return -1;
 	}
 
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
-		free(str);
 		return -1;
 	}
-	free(str);
 	return 0;
 }
 
 static int
 iscsi_login_add_sessiontype(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send TargetName during opneg or the first leg of secneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG
@@ -146,10 +138,10 @@ iscsi_login_add_sessiontype(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 
 	switch (iscsi->session_type) {
 	case ISCSI_SESSION_DISCOVERY:
-		str = (char *)"SessionType=Discovery";
+		strncpy(str,"SessionType=Discovery",MAX_STRING_SIZE);
 		break;
 	case ISCSI_SESSION_NORMAL:
-		str = (char *)"SessionType=Normal";
+		strncpy(str,"SessionType=Normal",MAX_STRING_SIZE);
 		break;
 	default:
 		iscsi_set_error(iscsi, "Can not handle sessions %d yet.",
@@ -168,7 +160,7 @@ iscsi_login_add_sessiontype(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 static int
 iscsi_login_add_headerdigest(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send HeaderDigest during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
@@ -177,16 +169,16 @@ iscsi_login_add_headerdigest(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 
 	switch (iscsi->want_header_digest) {
 	case ISCSI_HEADER_DIGEST_NONE:
-		str = (char *)"HeaderDigest=None";
+		strncpy(str,"HeaderDigest=None",MAX_STRING_SIZE);
 		break;
 	case ISCSI_HEADER_DIGEST_NONE_CRC32C:
-		str = (char *)"HeaderDigest=None,CRC32C";
+		strncpy(str,"HeaderDigest=None,CRC32C",MAX_STRING_SIZE);
 		break;
 	case ISCSI_HEADER_DIGEST_CRC32C_NONE:
-		str = (char *)"HeaderDigest=CRC32C,None";
+		strncpy(str,"HeaderDigest=CRC32C,None",MAX_STRING_SIZE);
 		break;
 	case ISCSI_HEADER_DIGEST_CRC32C:
-		str = (char *)"HeaderDigest=CRC32C";
+		strncpy(str,"HeaderDigest=CRC32C",MAX_STRING_SIZE);
 		break;
 	default:
 		iscsi_set_error(iscsi, "invalid header digest value");
@@ -205,14 +197,14 @@ iscsi_login_add_headerdigest(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 static int
 iscsi_login_add_datadigest(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send DataDigest during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = (char *)"DataDigest=None";
+	strncpy(str,"DataDigest=None",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -225,32 +217,24 @@ iscsi_login_add_datadigest(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 static int
 iscsi_login_add_initialr2t(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send InitialR2T during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = malloc(1024);
-#if defined(WIN32)
-	if (_snprintf_s(str, 1024, 1024, "InitialR2T=%s", iscsi->want_initial_r2t == ISCSI_INITIAL_R2T_NO ?
-#else
-	if (snprintf(str, 1024, "InitialR2T=%s", iscsi->want_initial_r2t == ISCSI_INITIAL_R2T_NO ?
-#endif
+	if (snprintf(str, MAX_STRING_SIZE, "InitialR2T=%s", iscsi->want_initial_r2t == ISCSI_INITIAL_R2T_NO ?
 		       "No" : "Yes") == -1) {
 		iscsi_set_error(iscsi, "Out-of-memory: aprintf failed.");
-		free(str);
 		return -1;
 	}
 
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
-		free(str);
 		return -1;
 	}
-	free(str);
 
 	return 0;
 }
@@ -258,32 +242,24 @@ iscsi_login_add_initialr2t(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 static int
 iscsi_login_add_immediatedata(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send ImmediateData during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = malloc(1024);
-#if defined(WIN32)
-	if (_snprintf_s(str, 1024, 1024, "ImmediateData=%s", iscsi->want_immediate_data == ISCSI_IMMEDIATE_DATA_NO ?
-#else
-	if (snprintf(str, 1024, "ImmediateData=%s", iscsi->want_immediate_data == ISCSI_IMMEDIATE_DATA_NO ?
-#endif
+	if (snprintf(str, MAX_STRING_SIZE, "ImmediateData=%s", iscsi->want_immediate_data == ISCSI_IMMEDIATE_DATA_NO ?
 		       "No" : "Yes") == -1) {
 		iscsi_set_error(iscsi, "Out-of-memory: aprintf failed.");
-		free(str);
 		return -1;
 	}
 
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
-		free(str);
 		return -1;
 	}
-	free(str);
 
 	return 0;
 }
@@ -291,107 +267,83 @@ iscsi_login_add_immediatedata(struct iscsi_context *iscsi, struct iscsi_pdu *pdu
 static int
 iscsi_login_add_maxburstlength(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send MaxBurstLength during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = malloc(1024);
-#if defined(WIN32)
-	if (_snprintf_s(str, 1024, 1024, "MaxBurstLength=%d", iscsi->max_burst_length) == -1) {
-#else
-	if (snprintf(str, 1024, "MaxBurstLength=%d", iscsi->max_burst_length) == -1) {
-#endif
+	if (snprintf(str, MAX_STRING_SIZE, "MaxBurstLength=%d", iscsi->max_burst_length) == -1) {
 		iscsi_set_error(iscsi, "Out-of-memory: aprintf failed.");
-		free(str);
 		return -1;
 	}
 
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
-		free(str);
 		return -1;
 	}
-	free(str);
 	return 0;
 }
 
 static int
 iscsi_login_add_firstburstlength(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send FirstBurstLength during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = malloc(1024);
-#if defined(WIN32)
-	if (_snprintf_s(str, 1024, 1024, "FirstBurstLength=%d", iscsi->first_burst_length) == -1) {
-#else
-	if (snprintf(str, 1024, "FirstBurstLength=%d", iscsi->first_burst_length) == -1) {
-#endif
+	if (snprintf(str, MAX_STRING_SIZE, "FirstBurstLength=%d", iscsi->first_burst_length) == -1) {
 		iscsi_set_error(iscsi, "Out-of-memory: aprintf failed.");
-		free(str);
 		return -1;
 	}
 
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
-		free(str);
 		return -1;
 	}
-	free(str);
 	return 0;
 }
 
 static int
 iscsi_login_add_maxrecvdatasegmentlength(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send MaxRecvDataSegmentLength during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = malloc(1024);
-#if defined(WIN32)
-	if (_snprintf_s(str, 1024, 1024, "MaxRecvDataSegmentLength=%d", iscsi->initiator_max_recv_data_segment_length) == -1) {
-#else
-	if (snprintf(str, 1024, "MaxRecvDataSegmentLength=%d", iscsi->initiator_max_recv_data_segment_length) == -1) {
-#endif
+	if (snprintf(str, MAX_STRING_SIZE, "MaxRecvDataSegmentLength=%d", iscsi->initiator_max_recv_data_segment_length) == -1) {
 		iscsi_set_error(iscsi, "Out-of-memory: aprintf failed.");
-		free(str);
 		return -1;
 	}
 
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
-		free(str);
 		return -1;
 	}
-	free(str);
 	return 0;
 }
 
 static int
 iscsi_login_add_datapduinorder(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send DataPduInOrder during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = (char *)"DataPDUInOrder=Yes";
+	strncpy(str,"DataPDUInOrder=Yes",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -404,14 +356,14 @@ iscsi_login_add_datapduinorder(struct iscsi_context *iscsi, struct iscsi_pdu *pd
 static int
 iscsi_login_add_defaulttime2wait(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send DefaultTime2Wait during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = (char *)"DefaultTime2Wait=2";
+	strncpy(str,"DefaultTime2Wait=2",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -424,14 +376,14 @@ iscsi_login_add_defaulttime2wait(struct iscsi_context *iscsi, struct iscsi_pdu *
 static int
 iscsi_login_add_defaulttime2retain(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send DefaultTime2Retain during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = (char *)"DefaultTime2Retain=0";
+	strncpy(str,"DefaultTime2Retain=0",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -444,14 +396,14 @@ iscsi_login_add_defaulttime2retain(struct iscsi_context *iscsi, struct iscsi_pdu
 static int
 iscsi_login_add_ifmarker(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send IFMarker during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = (char *)"IFMarker=No";
+	strncpy(str,"IFMarker=No",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -464,14 +416,14 @@ iscsi_login_add_ifmarker(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 static int
 iscsi_login_add_ofmarker(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send OFMarker during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = (char *)"OFMarker=No";
+	strncpy(str,"OFMarker=No",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -484,14 +436,14 @@ iscsi_login_add_ofmarker(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 static int
 iscsi_login_add_maxconnections(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send MaxConnections during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = (char *)"MaxConnections=1";
+	strncpy(str,"MaxConnections=1",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -504,14 +456,14 @@ iscsi_login_add_maxconnections(struct iscsi_context *iscsi, struct iscsi_pdu *pd
 static int
 iscsi_login_add_maxoutstandingr2t(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send MaxOutstandingR2T during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = (char *)"MaxOutstandingR2T=1";
+	strncpy(str,"MaxOutstandingR2T=1",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -524,14 +476,14 @@ iscsi_login_add_maxoutstandingr2t(struct iscsi_context *iscsi, struct iscsi_pdu
 static int
 iscsi_login_add_errorrecoverylevel(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send ErrorRecoveryLevel during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = (char *)"ErrorRecoveryLevel=0";
+	strncpy(str,"ErrorRecoveryLevel=0",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -544,14 +496,14 @@ iscsi_login_add_errorrecoverylevel(struct iscsi_context *iscsi, struct iscsi_pdu
 static int
 iscsi_login_add_datasequenceinorder(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	/* We only send DataSequenceInOrder during opneg */
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_OPNEG) {
 		return 0;
 	}
 
-	str = (char *)"DataSequenceInOrder=Yes";
+	strncpy(str,"DataSequenceInOrder=Yes",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -564,14 +516,14 @@ iscsi_login_add_datasequenceinorder(struct iscsi_context *iscsi, struct iscsi_pd
 static int
 iscsi_login_add_authmethod(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_SECNEG
 	|| iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_OFFER_CHAP) {
 		return 0;
 	}
 
-	str = (char *)"AuthMethod=CHAP,None";
+	strncpy(str,"AuthMethod=CHAP,None",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -584,14 +536,14 @@ iscsi_login_add_authmethod(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 static int
 iscsi_login_add_authalgorithm(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_SECNEG
 	|| iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_SELECT_ALGORITHM) {
 		return 0;
 	}
 
-	str = (char *)"CHAP_A=5";
+	strncpy(str,"CHAP_A=5",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str)+1)
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -604,14 +556,14 @@ iscsi_login_add_authalgorithm(struct iscsi_context *iscsi, struct iscsi_pdu *pdu
 static int
 iscsi_login_add_chap_username(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
 
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_SECNEG
 	|| iscsi->secneg_phase != ISCSI_LOGIN_SECNEG_PHASE_SEND_RESPONSE) {
 		return 0;
 	}
 
-	str = (char *)"CHAP_N=";
+	strncpy(str,"CHAP_N=",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str))
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -639,6 +591,7 @@ h2i(int h)
 	}
 	return h - '0';
 }
+
 static int
 i2h(int i)
 {
@@ -649,13 +602,49 @@ i2h(int i)
 	return i + '0';
 }
 
+#ifndef HAVE_LIBGCRYPT
+typedef struct MD5Context *gcry_md_hd_t;
+#define gcry_md_write MD5Update
+#define GCRY_MD_MD5 1
+
+static void gcry_md_open(gcry_md_hd_t *hd, int algo, unsigned int flags)
+{
+	assert(algo == GCRY_MD_MD5 && flags == 0);
+	*hd = malloc(sizeof(struct MD5Context));
+	if (*hd) {
+		MD5Init(*hd);
+	}
+}
+
+static void gcry_md_putc(gcry_md_hd_t h, unsigned char c)
+{
+	MD5Update(h, &c, 1);
+}
+
+static char *gcry_md_read(gcry_md_hd_t h, int algo)
+{
+	unsigned char digest[16];
+	assert(algo == 0 || algo == GCRY_MD_MD5);
+
+	MD5Final(digest, h);
+	return memcpy(h->buf, digest, sizeof(digest));
+}
+
+static void gcry_md_close(gcry_md_hd_t h)
+{
+	memset(h, 0, sizeof(*h));
+	free(h);
+}
+#endif
+
 static int
 iscsi_login_add_chap_response(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	char *str;
+	char str[MAX_STRING_SIZE+1];
+	char * strp;
 	unsigned char c, cc[2];
 	unsigned char digest[16];
-	struct MD5Context ctx;
+	gcry_md_hd_t ctx;
 	int i;
 
 	if (iscsi->current_phase != ISCSI_PDU_LOGIN_CSG_SECNEG
@@ -663,23 +652,29 @@ iscsi_login_add_chap_response(struct iscsi_context *iscsi, struct iscsi_pdu *pdu
 		return 0;
 	}
 
-	if (iscsi->chap_c == NULL) {
+	gcry_md_open(&ctx, GCRY_MD_MD5, 0);
+	if (!ctx) {
+		iscsi_set_error(iscsi, "Cannot create MD5 algorithm");
+		return -1;
+	}
+
+	if (!iscsi->chap_c[0]) {
 		iscsi_set_error(iscsi, "No CHAP challenge found");
 		return -1;
 	}
-	MD5Init(&ctx);
-	c = iscsi->chap_i;
-	MD5Update(&ctx, &c, 1);
-	MD5Update(&ctx, (unsigned char *)iscsi->passwd, strlen(iscsi->passwd));
-	str = iscsi->chap_c;
-	while (*str != 0) {
-		c = (h2i(str[0]) << 4) | h2i(str[1]);
-		str += 2;
-		MD5Update(&ctx, &c, 1);
-	}
-	MD5Final(digest, &ctx);
+	gcry_md_putc(ctx, iscsi->chap_i);
+	gcry_md_write(ctx, (unsigned char *)iscsi->passwd, strlen(iscsi->passwd));
 
-	str = (char *)"CHAP_R=0x";
+	strp = iscsi->chap_c;
+	while (*strp != 0) {
+		c = (h2i(strp[0]) << 4) | h2i(strp[1]);
+		strp += 2;
+		gcry_md_putc(ctx, c);
+	}
+	memcpy(digest, gcry_md_read(ctx, 0), sizeof(digest));
+	gcry_md_close(ctx);
+
+	strncpy(str,"CHAP_R=0x",MAX_STRING_SIZE);
 	if (iscsi_pdu_add_data(iscsi, pdu, (unsigned char *)str, strlen(str))
 	    != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: pdu add data failed.");
@@ -735,8 +730,10 @@ iscsi_login_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 		return -1;
 	}
 
-	pdu = iscsi_allocate_pdu(iscsi, ISCSI_PDU_LOGIN_REQUEST,
-				 ISCSI_PDU_LOGIN_RESPONSE);
+	pdu = iscsi_allocate_pdu_with_itt_flags(iscsi,
+				ISCSI_PDU_LOGIN_REQUEST,
+				ISCSI_PDU_LOGIN_RESPONSE,
+				iscsi->itt, 0);
 	if (pdu == NULL) {
 		iscsi_set_error(iscsi, "Out-of-memory: Failed to allocate "
 				"login pdu.");
@@ -746,7 +743,7 @@ iscsi_login_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 	/* login request */
 	iscsi_pdu_set_immediate(pdu);
 
-	if (iscsi->user == NULL) {
+	if (!iscsi->user[0]) {
 		iscsi->current_phase = ISCSI_PDU_LOGIN_CSG_OPNEG;
 	}
 
@@ -783,7 +780,7 @@ iscsi_login_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 	}
 
 	/* optional alias */
-	if (iscsi->alias) {
+	if (iscsi->alias[0]) {
 		if (iscsi_login_add_alias(iscsi, pdu) != 0) {
 			iscsi_free_pdu(iscsi, pdu);
 			return -1;
@@ -968,13 +965,14 @@ int
 iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 			  struct iscsi_in_pdu *in)
 {
-	int status;
+	uint32_t status;
 	char *ptr = (char *)in->data;
 	int size = in->data_pos;
 
-	status = ntohs(*(uint16_t *)&in->hdr[36]);
+	status = scsi_get_uint16(&in->hdr[36]);
 
-	iscsi->statsn = ntohs(*(uint16_t *)&in->hdr[24]);
+	iscsi_adjust_statsn(iscsi, in);
+	iscsi_adjust_maxexpcmdsn(iscsi, in);
 
 	/* XXX here we should parse the data returned in case the target
 	 * renegotiated some some parameters.
@@ -983,44 +981,34 @@ iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 	 */
 
 	while (size > 0) {
+		char *end;
 		int len;
 
-		len = strlen(ptr);
-
-		if (len == 0) {
-			break;
-		}
-
-		if (len > size) {
-			iscsi_set_error(iscsi, "len > size when parsing "
-					"login data %d>%d", len, size);
+		end = memchr(ptr, 0, size);
+		if (end == NULL) {
+			iscsi_set_error(iscsi, "NUL not found after offset %ld "
+					"when parsing login data",
+					(long)((unsigned char *)ptr - in->data));
 			pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
 				      pdu->private_data);
 			return -1;
 		}
 
+		len = end - ptr;
+		if (len == 0) {
+			break;
+		}
+
 		/* parse the strings */
 		if (!strncmp(ptr, "TargetAddress=", 14)) {
-			free(discard_const(iscsi->target_address));
-			iscsi->target_address = strdup(ptr+14);
-			if (iscsi->target_address == NULL) {
-				iscsi_set_error(iscsi, "Failed to allocate"
-						" target address");
-				pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
-					pdu->private_data);
-				return -1;
-			}
+			strncpy(iscsi->target_address,ptr+14,MAX_STRING_SIZE);
 		}
 
 		if (!strncmp(ptr, "HeaderDigest=", 13)) {
 			if (!strcmp(ptr + 13, "CRC32C")) {
-				iscsi->header_digest
-				  = ISCSI_HEADER_DIGEST_CRC32C;
 				iscsi->want_header_digest
 				  = ISCSI_HEADER_DIGEST_CRC32C;
 			} else {
-				iscsi->header_digest
-				  = ISCSI_HEADER_DIGEST_NONE;
 				iscsi->want_header_digest
 				  = ISCSI_HEADER_DIGEST_NONE;
 			}
@@ -1041,6 +1029,12 @@ iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 		if (!strncmp(ptr, "ImmediateData=", 14)) {
 			if (!strcmp(ptr + 14, "No")) {
 				iscsi->use_immediate_data = ISCSI_IMMEDIATE_DATA_NO;
+			} else if (iscsi->want_immediate_data == ISCSI_IMMEDIATE_DATA_NO) {
+				/* If we negotiated NO, it doesnt matter what
+				 * the target said. ImmediateData is NO.
+				 */
+				iscsi->use_immediate_data = ISCSI_IMMEDIATE_DATA_NO;
+
 			} else {
 				iscsi->use_immediate_data = ISCSI_IMMEDIATE_DATA_YES;
 			}
@@ -1071,13 +1065,7 @@ iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 		}
 
 		if (!strncmp(ptr, "CHAP_C=0x", 9)) {
-			free(iscsi->chap_c);
-			iscsi->chap_c = strdup(ptr+9);
-			if (iscsi->chap_c == NULL) {
-				iscsi_set_error(iscsi, "Out-of-memory: Failed to strdup CHAP challenge.");
-				pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL, pdu->private_data);
-				return -1;
-			}
+			strncpy(iscsi->chap_c,ptr+9,MAX_STRING_SIZE);
 			iscsi->secneg_phase = ISCSI_LOGIN_SECNEG_PHASE_SEND_RESPONSE;
 		}
 
@@ -1085,6 +1073,13 @@ iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 		size -= len + 1;
 	}
 
+	if (status == SCSI_STATUS_REDIRECT && iscsi->target_address[0]) {
+		ISCSI_LOG(iscsi, 2, "target requests redirect to %s",iscsi->target_address);
+		pdu->callback(iscsi, SCSI_STATUS_REDIRECT, NULL,
+				  pdu->private_data);
+		return 0;
+	}
+
 	if (status != 0) {
 		iscsi_set_error(iscsi, "Failed to log in to target. Status: %s(%d)",
 				       login_error_str(status), status);
@@ -1100,6 +1095,9 @@ iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 	if ((in->hdr[1] & ISCSI_PDU_LOGIN_TRANSIT)
 	&& (in->hdr[1] & ISCSI_PDU_LOGIN_NSG_FF) == ISCSI_PDU_LOGIN_NSG_FF) {
 		iscsi->is_loggedin = 1;
+		iscsi_itt_post_increment(iscsi);
+		iscsi->header_digest  = iscsi->want_header_digest;
+		ISCSI_LOG(iscsi, 2, "login successful");
 		pdu->callback(iscsi, SCSI_STATUS_GOOD, NULL, pdu->private_data);
 	} else {
 		if (iscsi_login_async(iscsi, pdu->callback, pdu->private_data) != 0) {
@@ -1111,10 +1109,9 @@ iscsi_process_login_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 	return 0;
 }
 
-
 int
-iscsi_logout_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
-		   void *private_data)
+iscsi_logout_async_internal(struct iscsi_context *iscsi, iscsi_command_cb cb,
+		   void *private_data, uint32_t flags)
 {
 	struct iscsi_pdu *pdu;
 
@@ -1148,6 +1145,7 @@ iscsi_logout_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 
 	pdu->callback     = cb;
 	pdu->private_data = private_data;
+	pdu->flags |= ISCSI_PDU_CORK_WHEN_SENT | flags;
 
 	if (iscsi_queue_pdu(iscsi, pdu) != 0) {
 		iscsi_set_error(iscsi, "Out-of-memory: failed to queue iscsi "
@@ -1160,9 +1158,18 @@ iscsi_logout_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 }
 
 int
-iscsi_process_logout_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
-struct iscsi_in_pdu *in _U_)
+iscsi_logout_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
+		   void *private_data)
 {
+	return iscsi_logout_async_internal(iscsi, cb, private_data, 0);
+}
+
+int
+iscsi_process_logout_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
+struct iscsi_in_pdu *in)
+{
+	iscsi_adjust_maxexpcmdsn(iscsi, in);
+
 	iscsi->is_loggedin = 0;
 	pdu->callback(iscsi, SCSI_STATUS_GOOD, NULL, pdu->private_data);
 
