@@ -14,12 +14,22 @@
    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, see <http://www.gnu.org/licenses/>.
 */
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
+#ifdef HAVE_SYS_TYPES_H
+#include <sys/types.h>
+#endif
+
+#ifdef HAVE_ARPA_INET_H
+#include <arpa/inet.h>
+#endif
 
 #if defined(WIN32)
 #include <winsock2.h>
 #else
 #include <strings.h>
-#include <arpa/inet.h>
 #endif
 
 #include <stdio.h>
@@ -30,28 +40,89 @@
 #include "scsi-lowlevel.h"
 #include "slist.h"
 
+/* This adds 32-bit serial comparision as defined in RFC1982.
+ * It returns 0 for equality, 1 if s1 is greater than s2 and
+ * -1 if s1 is less than s2. According to RFC1982 section 3.2
+ * there are rare cases where the result of the comparision is
+ * undefined e.g. when s1 = 0 and s2=2^31. This cases should
+ * not happen in iSCSI protocol.
+ */
+int
+iscsi_serial32_compare(uint32_t s1, uint32_t s2) {
+	if (s1 == s2) return 0;
+	if (s1 < s2 && s2-s1 < (uint32_t)1<<31) return -1;
+	if (s1 > s2 && s1-s2 < (uint32_t)1<<31) return 1;
+	if (s1 > s2 && s1-s2 > (uint32_t)1<<31) return -1;
+	if (s1 < s2 && s2-s1 > (uint32_t)1<<31) return 1;
+	/* undefined result */
+	return -1;
+}
+
+uint32_t
+iscsi_itt_post_increment(struct iscsi_context *iscsi) {
+	uint32_t old_itt = iscsi->itt;
+	iscsi->itt++;
+	/* 0xffffffff is a reserved value */
+	if (iscsi->itt == 0xffffffff) {
+		iscsi->itt = 0;
+	}
+	return old_itt;
+}
+
+void iscsi_adjust_statsn(struct iscsi_context *iscsi, struct iscsi_in_pdu *in) {
+	uint32_t statsn = scsi_get_uint32(&in->hdr[24]);
+	uint32_t itt = scsi_get_uint32(&in->hdr[16]);
+	
+	if (itt == 0xffffffff) {
+		/* target will not increase statsn if itt == 0xffffffff */
+		statsn--;
+	}
+	
+	if (iscsi_serial32_compare(statsn, iscsi->statsn) > 0) {
+		iscsi->statsn = statsn;
+	}
+}
+
+void iscsi_adjust_maxexpcmdsn(struct iscsi_context *iscsi, struct iscsi_in_pdu *in) {
+	uint32_t maxcmdsn = scsi_get_uint32(&in->hdr[32]);
+	if (iscsi_serial32_compare(maxcmdsn, iscsi->maxcmdsn) > 0) {
+		iscsi->maxcmdsn = maxcmdsn;
+	}
+	uint32_t expcmdsn = scsi_get_uint32(&in->hdr[28]);
+	if (iscsi_serial32_compare(expcmdsn, iscsi->expcmdsn) > 0) {
+		iscsi->expcmdsn = expcmdsn;
+	}
+}
+
+void iscsi_dump_pdu_header(struct iscsi_context *iscsi, unsigned char *data) {
+	char dump[ISCSI_RAW_HEADER_SIZE*3+1]={0};
+	int i;
+	for (i=0;i<ISCSI_RAW_HEADER_SIZE;i++) {
+		snprintf(&dump[i * 3], 4, " %02x", data[i]);
+	}
+	ISCSI_LOG(iscsi, 0, "PDU header:%s",dump);
+}
+
 struct iscsi_pdu *
 iscsi_allocate_pdu_with_itt_flags(struct iscsi_context *iscsi, enum iscsi_opcode opcode,
 				  enum iscsi_opcode response_opcode, uint32_t itt, uint32_t flags)
 {
 	struct iscsi_pdu *pdu;
 
-	pdu = malloc(sizeof(struct iscsi_pdu));
+	pdu = iscsi_szmalloc(iscsi, sizeof(struct iscsi_pdu));
 	if (pdu == NULL) {
 		iscsi_set_error(iscsi, "failed to allocate pdu");
 		return NULL;
 	}
-	memset(pdu, 0, sizeof(struct iscsi_pdu));
 
 	pdu->outdata.size = ISCSI_HEADER_SIZE;
-	pdu->outdata.data = malloc(pdu->outdata.size);
+	pdu->outdata.data = iscsi_szmalloc(iscsi, pdu->outdata.size);
 
 	if (pdu->outdata.data == NULL) {
 		iscsi_set_error(iscsi, "failed to allocate pdu header");
-		free(pdu);
+		iscsi_free(iscsi, pdu);
 		return NULL;
 	}
-	memset(pdu->outdata.data, 0, pdu->outdata.size);
 
 	/* opcode */
 	pdu->outdata.data[0] = opcode;
@@ -76,11 +147,9 @@ struct iscsi_pdu *
 iscsi_allocate_pdu(struct iscsi_context *iscsi, enum iscsi_opcode opcode,
 		   enum iscsi_opcode response_opcode)
 {
-	return iscsi_allocate_pdu_with_itt_flags(iscsi, opcode, response_opcode, iscsi->itt++, 0);
+	return iscsi_allocate_pdu_with_itt_flags(iscsi, opcode, response_opcode, iscsi_itt_post_increment(iscsi), 0);
 }	
 
-
-
 void
 iscsi_free_pdu(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
@@ -89,18 +158,25 @@ iscsi_free_pdu(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 		return;
 	}
 
-	free(pdu->outdata.data);
+	if (pdu->outdata.size <= iscsi->smalloc_size) {
+		iscsi_sfree(iscsi, pdu->outdata.data);
+	} else {
+		iscsi_free(iscsi, pdu->outdata.data);
+	}
 	pdu->outdata.data = NULL;
 
-	free(pdu->indata.data);
+	if (pdu->indata.size <= iscsi->smalloc_size) {
+		iscsi_sfree(iscsi, pdu->indata.data);
+	} else {
+		iscsi_free(iscsi, pdu->indata.data);
+	}
 	pdu->indata.data = NULL;
 
-	if (pdu->scsi_cbdata) {
-		iscsi_free_scsi_cbdata(pdu->scsi_cbdata);
-		pdu->scsi_cbdata = NULL;
+	if (iscsi->outqueue_current == pdu) {
+		iscsi->outqueue_current = NULL;
 	}
 
-	free(pdu);
+	iscsi_sfree(iscsi, pdu);
 }
 
 
@@ -108,8 +184,7 @@ int
 iscsi_add_data(struct iscsi_context *iscsi, struct iscsi_data *data,
 	       unsigned char *dptr, int dsize, int pdualignment)
 {
-	int len, aligned;
-	unsigned char *buf;
+	size_t len, aligned;
 
 	if (dsize == 0) {
 		iscsi_set_error(iscsi, "Trying to append zero size data to "
@@ -118,31 +193,37 @@ iscsi_add_data(struct iscsi_context *iscsi, struct iscsi_data *data,
 	}
 
 	len = data->size + dsize;
+
 	aligned = len;
 	if (pdualignment) {
 		aligned = (aligned+3)&0xfffffffc;
 	}
-	buf = malloc(aligned);
-	if (buf == NULL) {
+
+	if (data->size == 0) {
+		if (aligned <= iscsi->smalloc_size) {
+			data->data = iscsi_szmalloc(iscsi, aligned);
+		} else {
+			data->data = iscsi_malloc(iscsi, aligned);
+		}
+	} else {
+		if (aligned > iscsi->smalloc_size) {
+			data->data = iscsi_realloc(iscsi, data->data, aligned);
+		}
+	}
+	if (data->data == NULL) {
 		iscsi_set_error(iscsi, "failed to allocate buffer for %d "
-				"bytes", len);
+				"bytes", (int) len);
 		return -1;
 	}
 
-	if (data->size > 0) {
-		memcpy(buf, data->data, data->size);
-	}
-	memcpy(buf + data->size, dptr, dsize);
+	memcpy(data->data + data->size, dptr, dsize);
+	data->size += dsize;
+
 	if (len != aligned) {
 		/* zero out any padding at the end */
-	  memset(buf+len, 0, aligned-len);
+		memset(data->data + len, 0, aligned - len);
 	}
 
-	free(data->data);
-
-	data->data  = buf;
-	data->size = len;
-
 	return 0;
 }
 
@@ -166,8 +247,8 @@ iscsi_pdu_add_data(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 	}
 
 	/* update data segment length */
-	*(uint32_t *)&pdu->outdata.data[4] = htonl(pdu->outdata.size
-						   - ISCSI_HEADER_SIZE);
+	scsi_set_uint32(&pdu->outdata.data[4], pdu->outdata.size
+					       - ISCSI_HEADER_SIZE);
 
 	return 0;
 }
@@ -177,12 +258,23 @@ iscsi_get_pdu_data_size(const unsigned char *hdr)
 {
 	int size;
 
-	size = (ntohl(*(uint32_t *)&hdr[4])&0x00ffffff);
-	size = (size+3)&0xfffffffc;
+	size = scsi_get_uint32(&hdr[4]) & 0x00ffffff;
 
 	return size;
 }
 
+
+int
+iscsi_get_pdu_padding_size(const unsigned char *hdr)
+{
+	int data_size, padded_size;
+
+	data_size = scsi_get_uint32(&hdr[4]) & 0x00ffffff;
+	padded_size = (data_size+3) & 0xfffffffc;
+
+	return padded_size - data_size;
+}
+
 enum iscsi_reject_reason {
 	ISCSI_REJECT_RESERVED                 = 0x01,
 	ISCSI_REJECT_DATA_DIGEST_ERROR        = 0x02,
@@ -234,14 +326,10 @@ int iscsi_process_target_nop_in(struct iscsi_context *iscsi,
 				struct iscsi_in_pdu *in)
 {
 	uint32_t ttt;
-	uint32_t statsn;
 
-	ttt = ntohl(*(uint32_t *)&in->hdr[20]);
+	ttt = scsi_get_uint32(&in->hdr[20]);
 
-	statsn = ntohl(*(uint32_t *)&in->hdr[24]);
-	if (statsn > iscsi->statsn) {
-		iscsi->statsn = statsn;
-	}
+	iscsi_adjust_statsn(iscsi, in);
 
 	/* if the server does not want a response */
 	if (ttt == 0xffffffff) {
@@ -254,12 +342,25 @@ int iscsi_process_target_nop_in(struct iscsi_context *iscsi,
 }
 
 
+static void iscsi_reconnect_after_logout(struct iscsi_context *iscsi, int status,
+                        void *command_data _U_, void *opaque _U_)
+{
+	if (status) {
+		ISCSI_LOG(iscsi, 1, "logout failed: %s", iscsi_get_error(iscsi));
+	} else {
+		ISCSI_LOG(iscsi, 2, "logout was successful");
+	}
+	iscsi->pending_reconnect = 1;
+}
+
+
 int iscsi_process_reject(struct iscsi_context *iscsi,
 				struct iscsi_in_pdu *in)
 {
 	int size = in->data_pos;
 	uint32_t itt;
 	struct iscsi_pdu *pdu;
+	uint8_t reason = in->hdr[2];
 
 	if (size < ISCSI_RAW_HEADER_SIZE) {
 		iscsi_set_error(iscsi, "size of REJECT payload is too small."
@@ -268,7 +369,19 @@ int iscsi_process_reject(struct iscsi_context *iscsi,
 		return -1;
 	}
 
-	itt = ntohl(*(uint32_t *)&in->data[16]);
+	if (reason == ISCSI_REJECT_WAITING_FOR_LOGOUT) {
+		ISCSI_LOG(iscsi, 1, "target rejects request with reason: %s",  iscsi_reject_reason_str(reason));
+		iscsi_logout_async_internal(iscsi, iscsi_reconnect_after_logout, NULL, ISCSI_PDU_DROP_ON_RECONNECT|ISCSI_PDU_URGENT_DELIVERY);
+		return 0;
+	}
+
+	iscsi_set_error(iscsi, "Request was rejected with reason: 0x%02x (%s)", reason, iscsi_reject_reason_str(reason));
+
+	itt = scsi_get_uint32(&in->data[16]);
+
+	if (iscsi->log_level > 1) {
+		iscsi_dump_pdu_header(iscsi, in->data);
+	}
 
 	for (pdu = iscsi->waitpdu; pdu; pdu = pdu->next) {
 		if (pdu->itt == itt) {
@@ -283,10 +396,12 @@ int iscsi_process_reject(struct iscsi_context *iscsi,
 		return -1;
 	}
 
-	pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
-			      pdu->private_data);
+	if (pdu->callback) {
+		pdu->callback(iscsi, SCSI_STATUS_ERROR, NULL,
+					pdu->private_data);
+	}
 
-	SLIST_REMOVE(&iscsi->waitpdu, pdu);
+	ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
 	iscsi_free_pdu(iscsi, pdu);
 	return 0;
 }
@@ -302,20 +417,43 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 
 	opcode = in->hdr[0] & 0x3f;
 	ahslen = in->hdr[4];
-	itt = ntohl(*(uint32_t *)&in->hdr[16]);
+	itt = scsi_get_uint32(&in->hdr[16]);
 
 	if (ahslen != 0) {
 		iscsi_set_error(iscsi, "cant handle expanded headers yet");
 		return -1;
 	}
 
-	if (opcode == ISCSI_PDU_REJECT) {
-		iscsi_set_error(iscsi, "Request was rejected with reason: 0x%02x (%s)", in->hdr[2], iscsi_reject_reason_str(in->hdr[2]));
-
-		if (iscsi_process_reject(iscsi, in) != 0) {
+	if (opcode == ISCSI_PDU_ASYNC_MSG) {
+		uint8_t event = in->hdr[36];
+		uint16_t param1 = scsi_get_uint16(&in->hdr[38]); 
+		uint16_t param2 = scsi_get_uint16(&in->hdr[40]); 
+		uint16_t param3 = scsi_get_uint16(&in->hdr[42]); 
+		switch (event) {
+		case 0x1:
+			ISCSI_LOG(iscsi, 2, "target requests logout within %u seconds", param3);
+			iscsi_logout_async_internal(iscsi, iscsi_reconnect_after_logout, NULL, ISCSI_PDU_DROP_ON_RECONNECT|ISCSI_PDU_URGENT_DELIVERY);
+			return 0;
+		case 0x2:
+			ISCSI_LOG(iscsi, 2, "target will drop this connection. Time2Wait is %u seconds", param2);
+			iscsi->last_reconnect = time(NULL) + param2;
+			return 0;
+		case 0x3:
+			ISCSI_LOG(iscsi, 2, "target will drop all connections of this session. Time2Wait is %u seconds", param2);
+			iscsi->last_reconnect = time(NULL) + param2;
+			return 0;
+		case 0x4:
+			ISCSI_LOG(iscsi, 2, "target requests parameter renogitiation.");
+			iscsi_logout_async_internal(iscsi, iscsi_reconnect_after_logout, NULL, ISCSI_PDU_DROP_ON_RECONNECT);
+			return 0;
+		default:
+			ISCSI_LOG(iscsi, 1, "unhandled async event %u: param1 %u param2 %u param3 %u", event, param1, param2, param3);
 			return -1;
 		}
-		return 0;
+	}
+
+	if (opcode == ISCSI_PDU_REJECT) {
+		return iscsi_process_reject(iscsi, in);
 	}
 
 	if (opcode == ISCSI_PDU_NOP_IN && itt == 0xffffffff) {
@@ -348,7 +486,7 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 		 */
 		if (opcode == ISCSI_PDU_R2T) {
 			expected_response = ISCSI_PDU_R2T;
-	        }
+		}
 
 		if (opcode != expected_response) {
 			iscsi_set_error(iscsi, "Got wrong opcode back for "
@@ -359,7 +497,7 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 		switch (opcode) {
 		case ISCSI_PDU_LOGIN_RESPONSE:
 			if (iscsi_process_login_reply(iscsi, pdu, in) != 0) {
-				SLIST_REMOVE(&iscsi->waitpdu, pdu);
+				ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
 				iscsi_free_pdu(iscsi, pdu);
 				iscsi_set_error(iscsi, "iscsi login reply "
 						"failed");
@@ -368,7 +506,7 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 			break;
 		case ISCSI_PDU_TEXT_RESPONSE:
 			if (iscsi_process_text_reply(iscsi, pdu, in) != 0) {
-				SLIST_REMOVE(&iscsi->waitpdu, pdu);
+				ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
 				iscsi_free_pdu(iscsi, pdu);
 				iscsi_set_error(iscsi, "iscsi text reply "
 						"failed");
@@ -377,7 +515,7 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 			break;
 		case ISCSI_PDU_LOGOUT_RESPONSE:
 			if (iscsi_process_logout_reply(iscsi, pdu, in) != 0) {
-				SLIST_REMOVE(&iscsi->waitpdu, pdu);
+				ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
 				iscsi_free_pdu(iscsi, pdu);
 				iscsi_set_error(iscsi, "iscsi logout reply "
 						"failed");
@@ -386,7 +524,7 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 			break;
 		case ISCSI_PDU_SCSI_RESPONSE:
 			if (iscsi_process_scsi_reply(iscsi, pdu, in) != 0) {
-				SLIST_REMOVE(&iscsi->waitpdu, pdu);
+				ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
 				iscsi_free_pdu(iscsi, pdu);
 				iscsi_set_error(iscsi, "iscsi response reply "
 						"failed");
@@ -396,7 +534,7 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 		case ISCSI_PDU_DATA_IN:
 			if (iscsi_process_scsi_data_in(iscsi, pdu, in,
 						       &is_finished) != 0) {
-				SLIST_REMOVE(&iscsi->waitpdu, pdu);
+				ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
 				iscsi_free_pdu(iscsi, pdu);
 				iscsi_set_error(iscsi, "iscsi data in "
 						"failed");
@@ -405,7 +543,7 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 			break;
 		case ISCSI_PDU_NOP_IN:
 			if (iscsi_process_nop_out_reply(iscsi, pdu, in) != 0) {
-				SLIST_REMOVE(&iscsi->waitpdu, pdu);
+				ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
 				iscsi_free_pdu(iscsi, pdu);
 				iscsi_set_error(iscsi, "iscsi nop-in failed");
 				return -1;
@@ -414,7 +552,7 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 		case ISCSI_PDU_SCSI_TASK_MANAGEMENT_RESPONSE:
 			if (iscsi_process_task_mgmt_reply(iscsi, pdu,
 							  in) != 0) {
-				SLIST_REMOVE(&iscsi->waitpdu, pdu);
+				ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
 				iscsi_free_pdu(iscsi, pdu);
 				iscsi_set_error(iscsi, "iscsi task-mgmt failed");
 				return -1;
@@ -422,7 +560,7 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 			break;
 		case ISCSI_PDU_R2T:
 			if (iscsi_process_r2t(iscsi, pdu, in) != 0) {
-				SLIST_REMOVE(&iscsi->waitpdu, pdu);
+				ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
 				iscsi_free_pdu(iscsi, pdu);
 				iscsi_set_error(iscsi, "iscsi r2t "
 						"failed");
@@ -431,13 +569,13 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 			is_finished = 0;
 			break;
 		default:
-			iscsi_set_error(iscsi, "Dont know how to handle "
+			iscsi_set_error(iscsi, "Don't know how to handle "
 					"opcode 0x%02x", opcode);
 			return -1;
 		}
 
 		if (is_finished) {
-			SLIST_REMOVE(&iscsi->waitpdu, pdu);
+			ISCSI_LIST_REMOVE(&iscsi->waitpdu, pdu);
 			iscsi_free_pdu(iscsi, pdu);
 		}
 		return 0;
@@ -449,13 +587,13 @@ iscsi_process_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 void
 iscsi_pdu_set_itt(struct iscsi_pdu *pdu, uint32_t itt)
 {
-	*(uint32_t *)&pdu->outdata.data[16] = htonl(itt);
+	scsi_set_uint32(&pdu->outdata.data[16], itt);
 }
 
 void
 iscsi_pdu_set_ritt(struct iscsi_pdu *pdu, uint32_t ritt)
 {
-	*(uint32_t *)&pdu->outdata.data[20] = htonl(ritt);
+	scsi_set_uint32(&pdu->outdata.data[20], ritt);
 }
 
 void
@@ -473,54 +611,55 @@ iscsi_pdu_set_immediate(struct iscsi_pdu *pdu)
 void
 iscsi_pdu_set_ttt(struct iscsi_pdu *pdu, uint32_t ttt)
 {
-	*(uint32_t *)&pdu->outdata.data[20] = htonl(ttt);
+	scsi_set_uint32(&pdu->outdata.data[20], ttt);
 }
 
 void
 iscsi_pdu_set_cmdsn(struct iscsi_pdu *pdu, uint32_t cmdsn)
 {
-	*(uint32_t *)&pdu->outdata.data[24] = htonl(cmdsn);
+	scsi_set_uint32(&pdu->outdata.data[24], cmdsn);
 }
 
 void
 iscsi_pdu_set_rcmdsn(struct iscsi_pdu *pdu, uint32_t rcmdsn)
 {
-	*(uint32_t *)&pdu->outdata.data[32] = htonl(rcmdsn);
+	scsi_set_uint32(&pdu->outdata.data[32], rcmdsn);
 }
 
 void
 iscsi_pdu_set_datasn(struct iscsi_pdu *pdu, uint32_t datasn)
 {
-	*(uint32_t *)&pdu->outdata.data[36] = htonl(datasn);
+	scsi_set_uint32(&pdu->outdata.data[36], datasn);
 }
 
 void
 iscsi_pdu_set_expstatsn(struct iscsi_pdu *pdu, uint32_t expstatsnsn)
 {
-	*(uint32_t *)&pdu->outdata.data[28] = htonl(expstatsnsn);
+	scsi_set_uint32(&pdu->outdata.data[28], expstatsnsn);
 }
 
 void
 iscsi_pdu_set_bufferoffset(struct iscsi_pdu *pdu, uint32_t bufferoffset)
 {
-	*(uint32_t *)&pdu->outdata.data[40] = htonl(bufferoffset);
+	scsi_set_uint32(&pdu->outdata.data[40], bufferoffset);
 }
 
 void
 iscsi_pdu_set_cdb(struct iscsi_pdu *pdu, struct scsi_task *task)
 {
-  memset(&pdu->outdata.data[32], 0, 16);
+	memset(&pdu->outdata.data[32], 0, 16);
 	memcpy(&pdu->outdata.data[32], task->cdb, task->cdb_size);
 }
 
 void
 iscsi_pdu_set_lun(struct iscsi_pdu *pdu, uint32_t lun)
 {
-	pdu->outdata.data[9] = lun;
+	pdu->outdata.data[8] = lun >> 8;
+	pdu->outdata.data[9] = lun & 0xff;
 }
 
 void
 iscsi_pdu_set_expxferlen(struct iscsi_pdu *pdu, uint32_t expxferlen)
 {
-	*(uint32_t *)&pdu->outdata.data[20] = htonl(expxferlen);
+	scsi_set_uint32(&pdu->outdata.data[20], expxferlen);
 }
