@@ -14,9 +14,19 @@
    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, see <http://www.gnu.org/licenses/>.
 */
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
-#if defined(WIN32)
-#else
+#ifdef HAVE_SYS_TYPES_H
+#include <sys/types.h>
+#endif
+
+#ifdef HAVE_ARPA_INET_H
+#include <arpa/inet.h>
+#endif
+
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
@@ -86,6 +96,8 @@ iscsi_process_task_mgmt_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu
 
 	response = in->hdr[2];
 
+	iscsi_adjust_maxexpcmdsn(iscsi, in);
+
 	pdu->callback(iscsi, SCSI_STATUS_GOOD, &response, pdu->private_data);
 
 	return 0;
@@ -97,7 +109,7 @@ iscsi_task_mgmt_abort_task_async(struct iscsi_context *iscsi,
 		      struct scsi_task *task,
 		      iscsi_command_cb cb, void *private_data)
 {
-	iscsi_scsi_task_cancel(iscsi, task);
+	iscsi_scsi_cancel_task(iscsi, task);
 
 	return iscsi_task_mgmt_async(iscsi,
 		      task->lun, ISCSI_TM_ABORT_TASK,
